#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <sys/stat.h>
#include <string.h>
#include <sys/wait.h>

#include "dsi.h"
#include "sha1.h"

static dsi_es_context ctx;

char insavpath[256];
char inbanpath[256];
char cryptsavpath[256];
char cryptbanpath[256];
char savhashpath[256];
char banhashpath[256];
char sdkeypath[256];

void write_esblocks(unsigned char *buf, unsigned int size, FILE *fout)
{
	unsigned int tempsz = 0x20000;
	unsigned int pos = 0;
	unsigned char metablock[32];
	memset(metablock, 0, 32);

	while(pos<size)
	{
		if(size-pos<tempsz)tempsz = size-pos;

		dsi_es_encrypt(&ctx, &buf[pos], metablock, tempsz);
		fwrite(&buf[pos], 1, tempsz, fout);
		fwrite(metablock, 1, 32, fout);
		pos+= tempsz;
	}
}

int main(int argc, char **argv)
{
	struct stat filestat;
	FILE *fsavin, *fsavcrypt, *fkey;
	unsigned char *buf;
	int argi, i;
	int flags = 0;
	char *homepath;
	unsigned char sdkey[16];
	unsigned char hash[20];

	printf("dumprawsav v1.0\n");
	if(argc<3)
	{
		printf("Usage:\n");
		printf("dumprawsav <options>\n");
		printf("Options:\n");
		printf("--insav=<path> Input cleartext sav path.\n");
		printf("--inban=<path> Input cleartext banner path.\n");
		printf("--cryptsav=<path> Encrypt the sav specified by --insav into <path>.\n");
		printf("--cryptban=<path> Encrypt the banner specified by --inban into <path>.\n");
		printf("--hashsav=<path> Calc the cleartext sav SHA1 and write to <path>.\n");
		printf("--hashban=<path> Calc the cleartext ban SHA1 and write to <path>.\n");
		return 0;
	}

	memset(sdkeypath, 0, 256);
	homepath = getenv("HOME");
	if(homepath==NULL)snprintf(sdkeypath, 255, "sd_key");
	snprintf(sdkeypath, 255, "%s/.dsi/sd_key", homepath);

	memset(sdkey, 0, 16);
	fkey = fopen(sdkeypath, "rb");
	if(fkey==NULL)
	{
		printf("Failed to open %s\n", sdkeypath);
		return 1;
	}
	fread(sdkey, 1, 16, fkey);
	fclose(fkey);
	dsi_es_init(&ctx, sdkey);

	memset(insavpath, 0, 256);
	memset(inbanpath, 0, 256);
	memset(cryptsavpath, 0, 256);
	memset(cryptbanpath, 0, 256);
	memset(savhashpath, 0, 256);
	memset(banhashpath, 0, 256);
	for(argi=1; argi<argc; argi++)
	{
		if(strncmp(argv[argi], "--insav=", 8)==0)
		{
			flags |= 1<<0;
			strncpy(insavpath, &argv[argi][8], 0xff);
		}
		if(strncmp(argv[argi], "--inban=", 8)==0)
		{
			flags |= 1<<2;
			strncpy(inbanpath, &argv[argi][8], 0xff);
		}

		if(strncmp(argv[argi], "--cryptsav=", 11)==0)
		{
			flags |= 1<<1;
			strncpy(cryptsavpath, &argv[argi][11], 0xff);
		}
		if(strncmp(argv[argi], "--cryptban=", 11)==0)
		{
			flags |= 1<<3;
			strncpy(cryptbanpath, &argv[argi][11], 0xff);
		}

		if(strncmp(argv[argi], "--hashsav=", 10)==0)
		{
			strncpy(savhashpath, &argv[argi][10], 0xff);
		}
		if(strncmp(argv[argi], "--hashban=", 10)==0)
		{
			strncpy(banhashpath, &argv[argi][10], 0xff);
		}
	}

	if(flags & 0xf)
	{
		if((flags & 0x3)==0x3)
		{
			if(stat(insavpath, &filestat)==-1)
			{
				printf("Failed to stat %s\n", insavpath);
				return 2;
			}

			fsavcrypt = fopen(cryptsavpath, "wb");
			if(fsavcrypt==NULL)
			{
				printf("Failed to open %s\n", cryptsavpath);
				return 2;
			}
			fsavin = fopen(insavpath, "rb");
			if(fsavin==NULL)
			{
				printf("Failed to open %s\n", insavpath);
				return 2;
			}

			buf = (unsigned char*)malloc(filestat.st_size);
			if(buf==NULL)
			{
				printf("Failed to alloc mem for sav.\n");
				fclose(fsavcrypt);
				fclose(fsavin);
				return 3;
			}
			memset(buf, 0, filestat.st_size);

			fread(buf, 1, filestat.st_size, fsavin);
			fclose(fsavin);
			if(savhashpath[0])
			{
				SHA1(buf, filestat.st_size, hash);
				fsavin = fopen(savhashpath, "w");
				if(fsavin==NULL)
				{
					printf("Failed to open %s\n", savhashpath);
					fclose(fsavcrypt);
					free(buf);
					return 2;
				}
				for(i=0; i<20; i++)fprintf(fsavin, "%02x", hash[i]);
				fclose(fsavin);
			}

			write_esblocks(buf, filestat.st_size, fsavcrypt);
			fclose(fsavcrypt);

			

			free(buf);
		}
		if((flags & 0xc)==0xc)
		{
			if(stat(inbanpath, &filestat)==-1)
			{
				printf("Failed to stat %s\n", insavpath);
				return 2;
			}

			fsavcrypt = fopen(cryptbanpath, "wb");
			if(fsavcrypt==NULL)
			{
				printf("Failed to open %s\n", cryptbanpath);
				return 2;
			}
			fsavin = fopen(inbanpath, "rb");
			if(fsavin==NULL)
			{
				printf("Failed to open %s\n", insavpath);
				return 2;
			}

			buf = (unsigned char*)malloc(filestat.st_size);
			if(buf==NULL)
			{
				printf("Failed to alloc mem for sav.\n");
				fclose(fsavcrypt);
				fclose(fsavin);
				return 3;
			}
			memset(buf, 0, filestat.st_size);

			fread(buf, 1, filestat.st_size, fsavin);
			fclose(fsavin);
			if(savhashpath[0])
			{
				SHA1(buf, filestat.st_size, hash);
				fsavin = fopen(banhashpath, "w");
				if(fsavin==NULL)
				{
					printf("Failed to open %s\n", banhashpath);
					fclose(fsavcrypt);
					free(buf);
					return 2;
				}
				for(i=0; i<20; i++)fprintf(fsavin, "%02x", hash[i]);
				fclose(fsavin);
			}

			write_esblocks(buf, filestat.st_size, fsavcrypt);
			fclose(fsavcrypt);
			free(buf);
		}
	}

	return 0;
}

