#include "common.h"

void memset32(void* addr, int len);

void memset_arm7(u32 *buffer, int val, size_t len);
void memcpy_arm7(u32 *a, u32 *b, size_t len);
void memcpy8_arm7(u8 *a, u8 *b, size_t len);
void memset_addrs_arm7(u32 start, u32 end);

int main(void)
{
	int DSi_mode = 0;
	void (*func_ptr)(void);
	int i, reg;
	REG_IME = 0;
	while(BOOTFLAG!=1);
	if(*((vu8*)0x04004501))DSi_mode = 1;//I2CCNT, since 4000 is disabled for arm7 we use other regs to check dsimode.

	for(i=0; i<16; i++)//Reset sound.
	{
		SCHANNEL_CR(i) = 0;
		SCHANNEL_SOURCE(i) = 0;
		SCHANNEL_TIMER(i) = 0;
		SCHANNEL_LENGTH(i) = 0;
	}
	REG_SOUNDCNT = 0;
	REG_SOUNDBIAS = 0x0200; //Fix bad sounds after soft reset
	memset_arm7((void*)0x04000508, 0, 0x20);//Clear sound capture hw.

	for(i=0; i<4; i++)
	{
		DMA_CR(i) = 0;//Reset DMA.
		DMA_SRC(i) = 0;
		DMA_DEST(i) = 0;
		TIMER_CR(i) = 0;//Reset timers.
		TIMER_DATA(i) = 0;
		if(DSi_mode)
		{
			for(reg=0; reg<0x1c; reg+=4)*((u32*)(0x04004104 + ((i*0x1c)+reg))) = 0;//Reset NDMA.
		}
	}

	if(!DSi_mode)memset_addrs_arm7(0x03800000 - 0x8000, 0x03800000 + 0x10000);
	if(DSi_mode)
	{
		memset_addrs_arm7(0x03000000, 0x0380FFC0);
		memset_addrs_arm7(0x0380FFD0, 0x03800000 + 0x10000);
	}

	REG_IE = 0;
	REG_IF = ~0;
	REG_AUXIE = 0;
	REG_AUXIF = ~0;
	*((vu32*)(0x04000000-4)) = 0;//Clear IRQ stuff.
	*((vu32*)(0x04000000-8)) = ~0;
	REG_POWERCNT = 1;

	memcpy_arm7((u32*)NDSHEADER->arm7destination, (u32*)(BOOTLDR_NDS + NDSHEADER->arm7romOffset), NDSHEADER->arm7binarySize);
	if(DSi_mode && TWLNDSHEADER->arm7ibinarySize)memcpy_arm7((u32*)TWLNDSHEADER->arm7idestination, (u32*)(BOOTLDR_NDS + TWLNDSHEADER->arm7iromOffset), TWLNDSHEADER->arm7ibinarySize);//Load the 7i bin when used.

	BOOTFLAG = 2;
	while(REG_VCOUNT!=192);//Sync with Arm9 before jumping to the bin.
	while(REG_VCOUNT==192);
	BOOTFLAG = 3;
	
	func_ptr = (void*)NDSHEADER->arm7executeAddress;
	func_ptr();
	while(1);//Should never happen.
	return 0;
}

void memset_arm7(u32 *buffer, int val, size_t len)
{
	u32 *buf = buffer;
	while(len>0)
	{
		*buf = val;
		buf++;
		len-=4;
	}
}

void memcpy_arm7(u32 *a, u32 *b, size_t len)
{
	u32 *bufa = a;
	u32 *bufb = b;
	while(len>0)
	{
		*bufa = *bufb;
		bufa++;
		bufb++;
		len-=4;
	}
}

void memcpy8_arm7(u8 *a, u8 *b, size_t len)
{
	u8 *bufa = a;
	u8 *bufb = b;
	while(len>0)
	{
		*bufa = *bufb;
		bufa++;
		bufb++;
		len-=1;
	}
}

void memset8_arm7(u8 *buffer, int val, size_t len)
{
	u8 *buf = buffer;
	while(len>0)
	{
		*buf = val;
		buf++;
		len--;
	}
}

void memset_addrs_arm7(u32 start, u32 end)
{
	memset_arm7((u32*)start, 0, ((int)end - (int)start));
}

