/*
	fix checksum in guitarrock
	-- derpie

	build: gcc -o fix_guitar fix_guitar.c crc.c
*/

#include <stdio.h>
#include <stdint.h>

extern uint16_t crc16(uint16_t crc, const uint8_t *buffer, unsigned len);

uint16_t be16(unsigned char* p) {
	return (p[0] << 8) | p[1];
}

int main(int argc, char* argv[])
{
	if(argc < 2) {
		printf("Usage: %s Save.bin", argv[0]);
		return 0;
	}

	FILE* fp;
	unsigned len;
	uint16_t* buf;
	int needs_update = 0;
	int pos, i;
	unsigned char *buf8;
	
	fp = fopen(argv[1],"rb");
	
	if(fp == NULL) {
		printf("error: fopen\n");
		return 1;
	}
	
	fseek(fp, 0, SEEK_END);
	len = ftell(fp);
	rewind(fp);
	
	if((buf = (uint16_t*) malloc(len)) == NULL) {
		printf("error: malloc\n");
		return 1;
	}
	if(fread((char*) buf,len,1,fp) != 1) {
		printf("error: fread\n");
		return 1;
	}

	if(argc>2)
	{
		buf8 = (unsigned char*)buf;
		printf("Injecting padding...\n");
		for(pos=0x1984; pos<0x19AA; pos+=0x13)
		{
			for(i=0x1A8E - 0x1984; i>0; i--)buf8[pos+i] = buf8[pos+i-1];
			buf8[pos] = 0xF0;
		}
	}

	uint16_t crcA = crc16(0, (char*) buf+0x8, buf[3]+2);
	uint16_t crcAX = buf[0];
	
	if(crcA != crcAX) {
		printf("crcA needs updating! %x => %x\n", crcAX, crcA);
		buf[0] = crcA;
		needs_update=1;
	}
	else {
		printf("crcA match!\n");
	}
	
	uint16_t crcB = crc16(0, (char*) buf+0x18, buf[8+3]+2);
	uint16_t crcBX = buf[8];
	
	if(crcB != crcBX) {
		printf("crcB needs updating! %04x => %04x\n", crcBX, crcB);
		buf[8] = crcB;
		needs_update=1;
	}
	else {
		printf("crcB match!\n");
	}
	
	if(needs_update)
	{
		printf("Updating..\n");
		fp = freopen(argv[1],"wb+",fp);
		
		if(fp == NULL) {
			printf("fail: freopen\n");
			return 1;
		}
		if(fwrite(buf,len,1,fp) != 1) {
			printf("fail: fwrite\n");
			return 1;
		}
		fclose(fp);
		free(buf);
		printf("  OK!\n");
	}
	else
	{
		fclose(fp);
		free(buf);
		printf("Nothing to update..\n");
	}
	return 0;
}
